import java.util.ArrayList;
import java.util.List;
import java.util.Hashtable;
import java.util.Iterator;

public class Table {

	private String name;
	private List columnList;
	private List foreignKeysList;
	private Hashtable columnsByName;
	private Database tableParent;
	private int ID;
	private int pkID = 0;
	private int fkeysNumber;

	private Hashtable fkeyByName;

	/**
	 * Default Constructor
	 */
	public Table() {
		this(null);
	}
	/**
	 * Constructs a table object with a name
	 * 
	 * @param name
	 *                  table name
	 */
	public Table(String name) {
		this.name = name;
		columnList = new ArrayList();
		columnsByName = new Hashtable();
		foreignKeysList = new ArrayList(5);
		fkeyByName = new Hashtable();
		fkeysNumber = 0;
	}
	/**
	 * Get the name of the Table
	 */
	public String getName() {
		return name;
	}
	/**
	 * Set the name of the Table
	 */
	public void setName(String newName) {
		name = newName;
	}
	/**
	 * Set the ID of the Table
	 */
	public void setID(int newID) {
		ID = newID;
	}
	/**
	 * get the ID of the Table
	 */
	public int getID() {
		return ID;
	}

	/**
	 * Adds a new column to the column list and set the parent table of the column to the current table
	 * 
	 * @param col
	 *                  the column to add
	 */
	public void addColumn(Column col) {
		col.setTable(this);
		columnList.add(col);
		columnsByName.put(col.getName(), col);
	}

	/**
	 * Returns an Array containing all the columns in the table
	 */
	public Column[] getColumns() {
		int size = columnList.size();
		Column[] tbls = new Column[size];
		for (int i = 0; i < size; i++) {
			tbls[i] = (Column) columnList.get(i);
		}
		return tbls;
	}
	/**
	 * Utility method to get the number of columns in this table
	 */
	public int getNumColumns() {
		return columnList.size();
	}
	/**
	 * Returns a specified column.
	 * 
	 * @param name
	 *                  name of the column
	 * @return Return a Column object or null if it does not exist.
	 */
	public Column getColumn(String name) {
		return (Column) columnsByName.get(name);
	}
	/**
	 * Set the parent of the table
	 * 
	 * @param parent
	 *                  the parant database
	 */
	public void setDatabase(Database parent) {
		tableParent = parent;
	}
	/**
	 * Get the parent of the table
	 * 
	 * @return the parant database
	 */
	public Database getDatabase() {
		return tableParent;
	}

	/**
	 * Returns the collection of Columns which make up the single primary key for this table.
	 * 
	 * @return A list of the primary key parts.
	 */
	public List getPrimaryKey() {
		List pk = new ArrayList(columnList.size());

		Iterator iter = columnList.iterator();
		while (iter.hasNext()) {
			Column col = (Column) iter.next();
			if (col.isPrimaryKey()) {
				pk.add(col);
			}
		}
		return pk;
	}
	/**
	 * Determine whether this table has a primary key.
	 * 
	 * @return Whether this table has any primary key parts.
	 */
	public boolean hasPrimaryKey() {
		return (getPrimaryKey().size() > 0);
	}
	/**
	 * Add a FK to a table
	 */
	public void addForeignKey(String tableSource, String colSource, String tableForeign, String colForeign) {

		//internal fk name, useful to check if the fk already exists
		String fkeyName = "FK_" + tableSource + "_" + tableForeign;
		ForeignKey myfkey = getFKey(fkeyName);

		//check if it is a new Fk or if it already exists, just add new attributes
		if (myfkey != null) {
			myfkey.addReference(colSource, colForeign);
		} else {
			ForeignKey fkey = new ForeignKey(tableSource, tableForeign, fkeysNumber, ID);
			fkeysNumber++;
			foreignKeysList.add(fkey);
			fkey.addReference(colSource, colForeign);
			fkeyByName.put(fkey.getName(), fkey);
		}
	}
	public ForeignKey getFKey(String name) {
		return (ForeignKey) fkeyByName.get(name);
	}

	/**
	 * Returns a representation of this table.
	 * 
	 * @return representation of this table
	 */
	public String toSQL(boolean dropTables) throws Exception {
		StringBuffer result = new StringBuffer();
		result.append("-- \n");
		result.append("-- TABLE: ").append(name).append("\n");
		result.append("-- \n");
	    if (dropTables){
	        result.append("DROP TABLE ").append(name).append("\n@\n");
	    }					
		result.append("CREATE TABLE ").append(name).append("(").append("\n");
		if (columnList != null) {
			for (Iterator iter = columnList.iterator(); iter.hasNext();) {
				Column col = (Column) iter.next();
				result.append(col.toSQL(getPrimaryKey().size()));
			}
		}
		if (hasPrimaryKey()) {
			int counterComma = 0;
			result.append("     CONSTRAINT ");
			String pkIdent = "PK_" + ID;
			//max pk name length = 18
			if (pkIdent.length() > 18) {
				throw new Exception("PK Constraint name too long");
			} else {
				result.append(pkIdent);
			}
			result.append(" PRIMARY KEY (");

			for (Iterator iter = getPrimaryKey().iterator(); iter.hasNext();) {
				Column pkCol = (Column) iter.next();
				result.append(pkCol.getName());

				// check conditions in order to put the commas between the columns composing the pk :
				// (col1,col2,col3...)
				if (getPrimaryKey().size() > 1 && counterComma < getPrimaryKey().size() - 1) {
					result.append(",");
					counterComma++;
				}
			}
			result.append(")\n");
		}
		result.append(")\n").append("@\n\n");
		return result.toString();
	}
	/**
	 * Returns a representation of the foreign keys associated to this table.
	 * 
	 * @return representation of the fk associated to this table
	 */
	public String toSQLbis() throws Exception {
		StringBuffer result = new StringBuffer();

		if (!foreignKeysList.isEmpty()) {
			result.append("-- \n");
			result.append("-- TABLE: ").append(name).append("\n");
			result.append("-- \n");
			for (Iterator iter = foreignKeysList.iterator(); iter.hasNext();) {
				ForeignKey fkey = (ForeignKey) iter.next();
				result.append(fkey.toSQL());
				result.append("@\n");
			}
		}

		return result.toString();
	}

	//BETA TEST
	public void toDB(ConnectionDB conn,boolean dropTables) {
		StringBuffer result;		
		try {
		    if (dropTables){
		        result = new StringBuffer();
		        result.append("DROP TABLE ").append(name);
		        conn.Execute(result.toString());
		    }
		} catch (Exception ex) {
			System.err.println("[-] table.toDB (DROP).");
		} finally {
			
			try {
				result = new StringBuffer();
				result.append("CREATE TABLE ").append(name).append("(");
				if (columnList != null) {
					for (Iterator iter = columnList.iterator(); iter.hasNext();) {
						Column col = (Column) iter.next();
						result.append(col.toDB(getPrimaryKey().size()));
					}
				}
				if (hasPrimaryKey()) {
					int counterComma = 0;
					result.append(" CONSTRAINT");
					String pkIdent = " PK_" + ID;
					//max pk name length = 18
					if (pkIdent.length() > 18) {
						throw new Exception("PK Constraint name too long");
					} else {
						result.append(pkIdent);
					}
					result.append(" PRIMARY KEY (");
														
					for (Iterator iter = getPrimaryKey().iterator(); iter.hasNext();) {
						Column pkCol = (Column) iter.next();
						result.append(pkCol.getName());

						// check conditions in order to put the commas between the columns composing the pk :
						// (col1,col2,col3...)
						if (getPrimaryKey().size() > 1 && counterComma < getPrimaryKey().size() - 1) {
							result.append(",");
							counterComma++;
						}
					}
					result.append(")");
				}
				result.append(")");
				//create table
				//System.out.println("----------------->" + result);
				conn.Execute(result.toString());
			} catch (Exception ex) {
				System.err.println("[-] table.toDB (CREATE).");
			}
		}

	}

	public void toDBbis(ConnectionDB conn) throws Exception {
		if (!foreignKeysList.isEmpty()) {
			for (Iterator iter = foreignKeysList.iterator(); iter.hasNext();) {
				ForeignKey fkey = (ForeignKey) iter.next();
				conn.Execute(fkey.toDB());
			}
		}
	}

}
